import { RecordNotFoundError } from '@tomei/general';
import { AccountSystemEntity } from '../account-system-entity/account-system-entity';
import AccountModel from '../models/account.entity';
import { AccountRepository } from './account.repository';

// eslint-disable-next-line @typescript-eslint/no-var-requires
const getConfig = require('../config');

export default class Account extends AccountSystemEntity {
  AccountNo = 'New';
  ParentAccountNo? = '';
  Name = '';
  Description = '';
  AccountType = '';
  AccountSubtype = '';
  OwnerId = '';
  OwnerType = '';
  RelatedObjectId = '';
  RelatedObjectType = '';
  CreatedById = '';
  CreatedAt: Date;
  UpdatedById = '';
  UpdatedAt: Date;

  private static _RepositoryBase = new AccountRepository();

  private _DbTransaction: any;
  private _IsNewRecord = true;

  private get IsNewRecord() {
    return this._IsNewRecord;
  }

  private set IsNewRecord(record: boolean) {
    this._IsNewRecord = record;
  }

  /* AccountSystemEntity Implementation */
  get RepositoryBase() {
    return Account._RepositoryBase;
  }

  /* ObjectBase Implementation */
  get ObjectId() {
    return this.AccountNo;
  }

  get ObjectName() {
    return this.Name;
  }

  get TableName() {
    return 'finance_Account';
  }

  constructor(dbTransaction?: any, accountNo?: string) {
    super();
    if (dbTransaction) {
      this._DbTransaction = dbTransaction;
    }
    if (accountNo) {
      this.RepositoryBase.findOne({
        where: {
          AccountNo: accountNo,
        },
        transaction: dbTransaction,
      })
        .then((accountData) => {
          if (accountData) {
            this.IsNewRecord = false;
            this.CompanyId = accountData.CompanyId;
            this.ParentAccountNo = accountData.ParentAccountNo;
            this.Name = accountData.Name;
            this.Description = accountData.Description;
            this.AccountType = accountData.AccountType;
            this.AccountSubtype = accountData.AccountSubType;
            this.OwnerId = accountData.OwnerId;
            this.OwnerType = accountData.OwnerType;
            this.RelatedObjectId = accountData.RelatedObjectId;
            this.RelatedObjectType = accountData.RelatedObjectType;
            this.CreatedById = accountData.CreatedById;
            this.CreatedAt = accountData.CreatedAt;
            this.UpdatedById = accountData.UpdatedById;
            this.UpdatedAt = accountData.UpdatedAt;
            this.AccSystemRefId = accountData.AccSystemRefId;
            this.PostedToAccSystemYN = accountData.PostedToAccSystemYN;
            this.PostedById = accountData.PostedById;
            this.PostedDateTime = accountData.PostedDateTime;
          } else {
            const notFoundError = new RecordNotFoundError('No Record Found.');
            throw notFoundError;
          }
        })
        .catch((err) => {
          // tslint:disable-next-line:no-console
          console.log('Account Class constructor err: ', err);
        });
    }
  }

  private init(accountData: any) {
    this.IsNewRecord = false;
    this.CompanyId = accountData.CompanyId;
    this.ParentAccountNo = accountData.ParentAccountNo;
    this.Name = accountData.Name;
    this.Description = accountData.Description;
    this.AccountType = accountData.AccountType;
    this.AccountSubtype = accountData.AccountSubType;
    this.OwnerId = accountData.OwnerId;
    this.OwnerType = accountData.OwnerType;
    this.RelatedObjectId = accountData.RelatedObjectId;
    this.RelatedObjectType = accountData.RelatedObjectType;
    this.CreatedById = accountData.CreatedById;
    this.CreatedAt = accountData.CreatedAt;
    this.UpdatedById = accountData.UpdatedById;
    this.UpdatedAt = accountData.UpdatedAt;
    this.AccSystemRefId = accountData.AccSystemRefId;
    this.PostedToAccSystemYN = accountData.PostedToAccSystemYN;
    this.PostedById = accountData.PostedById;
    this.PostedDateTime = accountData.PostedDateTime;
  }

  static async initAccount(
    dbTransaction: any,
    accountNo: string,
  ): Promise<Account> {
    try {
      const accountData = await Account._RepositoryBase.findOne({
        where: {
          AccountNo: accountNo,
        },
        transaction: dbTransaction,
      });
      if (accountData) {
        const account = new Account(dbTransaction);
        account.init(accountData.get({ plain: true }));
        return account;
      } else {
        const notFoundError = new RecordNotFoundError('No Record Found.');
        throw notFoundError;
      }
    } catch (error) {
      console.log('Account Class constructor err: ', error);
      throw error;
    }
  }

  /**
   * Account static constructor
   *
   * Create the Customer Account Receivable (AR) and Customer Account Payable (AP) default accounts.
   * The static constructor is to run once each time the program loads and check if the default accounts have been created.
   * If it has not, it will create the accounts.
   */
  static initialize(): void {
    const dbTransaction = null;
    let defaultAccountReceivable: Account;
    let defaultAccountPayable: Account;

    // const config = getConfig();
    // config.systemConfig['EZC']

    try {
      defaultAccountReceivable = new Account(dbTransaction, 'EZC-AR');
    } catch (err) {
      defaultAccountReceivable = new Account(dbTransaction);
      defaultAccountReceivable.AccountNo = null;
      defaultAccountReceivable.AccountType = 'Account Receivable';
      defaultAccountReceivable.OwnerId = 'System';
      defaultAccountReceivable.OwnerType = '';
    }

    try {
      defaultAccountPayable = new Account(dbTransaction, 'EZC-AP');
    } catch (err) {
      defaultAccountPayable = new Account(dbTransaction);
      defaultAccountPayable.AccountNo = null;
      defaultAccountPayable.AccountType = 'Account Payable';
      defaultAccountPayable.OwnerId = 'System';
    }
  }

  validateAccountValue() {
    if (!this.AccSystemRefId) {
      throw new Error('AccSystemAccountId is required.');
    }
    if (!this.AccountType) {
      throw new Error('AccountType is required.');
    }
  }

  isParentAccountExists() {
    if (this.ParentAccountNo) return true;
    return false;
  }

  async save(
    AccSystemRefId: string,
    userId: string,
    dbTransaction?: any,
  ): Promise<AccountModel> {
    this.AccSystemRefId = AccSystemRefId;
    const data = await this.RepositoryBase.create(
      {
        AccountNo: this.AccountNo,
        AccSystemRefId: this.AccSystemRefId,
        Name: this.Name,
        Description: this.Description,
        AccountType: this.AccountType,
        AccountSubtype: this.AccountSubtype,
        OwnerId: this.OwnerId,
        OwnerType: this.OwnerType,
        RelatedObjectId: this.RelatedObjectId,
        RelatedObjectType: this.RelatedObjectType,
        CreatedAt: new Date(),
        CreatedById: userId,
      },
      dbTransaction,
    );
    return data;
  }
}
