import {
  IAddress,
  IRepositoryBase,
  IPerson,
  LoginUserBase,
  RecordNotFoundError,
} from '@tomei/general';
import FinanceCustomerModel from '../models/customer.entity';
import { AccountSystemEntity } from '../account-system-entity/account-system-entity';
import { FinanceCustomerRepository } from './finance-customer.repository';
import * as cuid from 'cuid';
import Account from '../account/account';
import { AccountRepository } from '../account/account.repository';
import type from '../helpers/typeof';
import * as FinanceDb from '../database';

// const AccountSystemEntity = import('../account-system-entity').then(({AccountSystemEntity}) => AccountSystemEntity);
export default abstract class FinanceCustomerBase
  extends AccountSystemEntity
  implements IPerson
{
  CustomerId = 'New';
  CustSystemCode = '';
  CustSystemRefId = '';

  protected abstract ParentARAccountNo: string;
  protected abstract ParentAPAccountNo: string;
  protected abstract ARAccountNo: string;
  protected abstract APAccountNo: string;

  abstract FullName: string;
  abstract IDNo: string;
  abstract IDType: string;
  abstract ContactNo: string;
  abstract Email: string;
  abstract DefaultAddress: IAddress;

  private _AccountReceivable: Account;
  private _AccountPayable: Account;

  private static _AccountRepository = new AccountRepository();

  // note: getDetails from spec is void type. Meaning that this probably needed a fix soon
  async getDetails(): Promise<{
    FullName: string;
    IDNo: string;
    IDType: string;
    Email: string;
    ContactNo: string;
  }> {
    return {
      FullName: this.FullName,
      IDNo: this.IDNo,
      IDType: this.IDType,
      Email: this.Email,
      ContactNo: this.ContactNo,
    };
  }

  protected static _FinanceCustomerRepository = new FinanceCustomerRepository();
  RepositoryBase: IRepositoryBase<any>;

  constructor(
    custSystemRefId?: string,
    custSystemCode?: string,
    sFinanceCompanyId?: string,
  ) {
    super();
    if (custSystemRefId && custSystemCode && sFinanceCompanyId) {
      const financeCustomerData =
        FinanceCustomerBase._FinanceCustomerRepository.findOne({
          where: {
            CompanyId: sFinanceCompanyId,
            CustSystemCode: custSystemCode,
            CustSystemRefId: custSystemRefId,
          },
        });
      if (financeCustomerData) {
        this.CompanyId = sFinanceCompanyId;
        this.CustSystemCode = custSystemCode;
        this.CustSystemRefId = custSystemRefId;
      } else {
        const notFoundError = new RecordNotFoundError('No Record Found');
        throw notFoundError;
      }
    }
  }

  public static async initCustomer<C extends FinanceCustomerBase>(
    this: new () => C,
    custSystemRefId: string,
    custSystemCode: string,
    sFinanceCompanyId: string,
  ): Promise<C> {
    const financeCustomerData =
      await FinanceCustomerBase._FinanceCustomerRepository.findOne({
        where: {
          CompanyId: sFinanceCompanyId,
          CustSystemCode: custSystemCode,
          CustSystemRefId: custSystemRefId,
        },
      });
    if (financeCustomerData) {
      const self = new this();
      self.CustomerId = financeCustomerData.CustomerId;
      self.CompanyId = sFinanceCompanyId;
      self.CustSystemCode = custSystemCode;
      self.CustSystemRefId = custSystemRefId;
      self.AccSystemRefId = financeCustomerData.AccSystemRefId;
      self.PostedToAccSystemYN = financeCustomerData.PostedToAccSystemYN;
      self.PostedById = financeCustomerData.PostedById;
      self.PostedDateTime = financeCustomerData.PostedDateTime;
      return self;
    } else {
      const notFoundError = new RecordNotFoundError('No Record Found');
      throw notFoundError;
    }
  }

  get AccountReceivable(): Promise<Account> {
    return new Promise((resolve, reject) => {
      let transaction;
      if (this._AccountReceivable) {
        resolve(this._AccountReceivable);
      } else {
        FinanceDb.getConnection()
          .transaction()
          .then((t) => {
            transaction = t;
            return FinanceCustomerBase._AccountRepository.findOne({
              where: {
                AccountType: 'Account Receivable',
                OwnerId: this.CustomerId,
                OwnerType: type(this),
              },
            });
          })
          .then((accountData) => {
            this._AccountReceivable = new Account(transaction);
            if (accountData) {
              this._AccountReceivable.AccountNo = accountData.AccountNo;
              this._AccountReceivable.ParentAccountNo =
                accountData.ParentAccountNo;
              this._AccountReceivable.AccountType = accountData.AccountType;
              this._AccountReceivable.OwnerId = accountData.OwnerId;
              this._AccountReceivable.OwnerType = accountData.OwnerType;
            } else {
              // this._AccountReceivable.AccountNo = `${this.CustSystemCode}-AR-123456789`;
              // this._AccountReceivable.ParentAccountNo = `${this.CustSystemCode}-AR`;
              this._AccountReceivable.AccountNo = this.ARAccountNo; //SystemInfo.SYSTEM_CODE + '-AR-' + '8 digit running no'
              this._AccountReceivable.ParentAccountNo = this.ParentARAccountNo; //SystemInfo.SYSTEM_CODE + '-AR'
              this._AccountReceivable.AccountType = 'Account Receivable';
              this._AccountReceivable.OwnerId = this.CustomerId;
              this._AccountReceivable.OwnerType = type(this);
            }
            resolve(this._AccountReceivable);
          })
          .catch((err) => {
            reject(err);
          });
      }
    });
  }

  get AccountPayable(): Promise<Account> {
    return new Promise((resolve, reject) => {
      let transaction;
      if (this._AccountPayable) {
        resolve(this._AccountPayable);
      } else {
        FinanceDb.getConnection()
          .transaction()
          .then((t) => {
            transaction = t;
            return FinanceCustomerBase._AccountRepository.findOne({
              where: {
                AccountType: 'Account Payable',
                OwnerId: this.CustomerId,
                OwnerType: type(this),
              },
            });
          })
          .then((accountData) => {
            this._AccountPayable = new Account(transaction);
            if (accountData) {
              this._AccountPayable.AccountNo = accountData.AccountNo;
              this._AccountPayable.ParentAccountNo =
                accountData.ParentAccountNo;
              this._AccountPayable.AccountType = accountData.AccountType;
              this._AccountPayable.OwnerId = accountData.OwnerId;
              this._AccountPayable.OwnerType = accountData.OwnerType;
            } else {
              // this._AccountPayable.AccountNo = `${this.CustSystemCode}-AP-123456789`;
              // this._AccountPayable.ParentAccountNo = `${this.CustSystemCode}-AP`;
              this._AccountPayable.AccountNo = this.APAccountNo; //SystemInfo.SYSTEM_CODE + '-AP-' + '8 digit running no'
              this._AccountPayable.ParentAccountNo = this.ParentAPAccountNo; //SystemInfo.SYSTEM_CODE + '-AP'
              this._AccountPayable.AccountType = 'Account Payable';
              this._AccountPayable.OwnerId = this.CustomerId;
              this._AccountPayable.OwnerType = type(this);
            }
            resolve(this._AccountPayable);
          })
          .catch((err) => {
            reject(err);
          });
      }
    });
  }

  /*
   * Get the billing address for a person.
   **/
  abstract getBillingAddress(params: any): Promise<IAddress>;

  init(person: LoginUserBase) {
    this.FullName = person.FullName;
    this.IDNo = person.IDNo;
    this.IDType = person.IDType;
    this.Email = person.Email;
    this.ContactNo = person.ContactNo;
    this.DefaultAddress = person.DefaultAddress;
  }

  async save(
    accSystemRefId: string,
    custSystemCode: string,
    custSystemRefId: string,
    dbTransaction?: any,
  ): Promise<FinanceCustomerModel> {
    this.CustSystemCode = custSystemCode;
    this.CustSystemRefId = custSystemRefId;
    this.AccSystemRefId = accSystemRefId;
    this.PostedToAccSystemYN = 'Y';
    const data = await FinanceCustomerBase._FinanceCustomerRepository.create(
      {
        CompanyId: this.CompanyId,
        CustomerId: cuid(),
        CustSystemCode: this.CustSystemCode,
        CustSystemRefId: this.CustSystemRefId,
        AccSystemRefId: this.AccSystemRefId,
        PostedToAccSystemYN: this.PostedToAccSystemYN,
      },
      { transaction: dbTransaction },
    );
    return data;
  }

  static async isFinanceCustomerExist(
    custSystemRefId: string,
    custSystemCode: string,
    sFinanceCompanyId: string,
  ): Promise<boolean> {
    const data = await FinanceCustomerBase._FinanceCustomerRepository.findAll({
      where: {
        CustSystemRefId: custSystemRefId,
        CustSystemCode: custSystemCode,
        CompanyId: sFinanceCompanyId,
      },
    });

    if (data.length > 0) {
      return true;
    }
    return false;
  }
}
