import * as util from 'util';
import * as fs from 'fs';
import * as puppeteer from 'puppeteer';
import { RecordNotFoundError } from '@tomei/general';
import {
  Medias,
  InternalMediaDto,
  CommonService as MediaCommonService,
  MediaType,
  MediasModel,
  MediasRepository,
} from '@tomei/media';
import { DocumentStatus, DocType } from '../enum';
import { IDocumentAttr } from './interfaces/document-attr.interface';
import { DocumentItemRepository } from './document-item.repository';
import DocumentItem from './document-item';
import { AccountSystemEntity } from '../account-system-entity/account-system-entity';
import { DocumentRepository } from './document.repository';
import FinanceCustomerBase from '../customer/customer';
import DocumentModel from '../models/document.entity';
import * as path from 'path';

export default class Document extends AccountSystemEntity {
  DocNo = 'New';
  private _DocType: DocType;
  private _DocDate = new Date();
  Currency = 'MYR';
  private _Amount = 0;
  Description = '';
  private _Status = DocumentStatus.UNPAID;
  IssuedById = '';
  IssuedToId = '';
  IssuedToType = '';
  RelatedObjectId = '';
  RelatedObjectType = '';
  CreatedById = '';
  CreatedAt: Date;
  UpdatedById = '';
  UpdatedAt: Date;
  UseAccSystemDocYN = 'N';

  private _DocHTMLFileMediaId = '';
  private _DocPDFFileMediaId = '';

  private static _RepositoryBase = new DocumentRepository();
  private static _DocumentItemRepository = new DocumentItemRepository();

  private _DbTransaction: any;
  private _IsNewRecord = true;
  private _DocumentItems = [];

  private static _MediaRepository = new MediasRepository(MediasModel);
  mediaCommonService: MediaCommonService;

  private get IsNewRecord() {
    return this._IsNewRecord;
  }

  private set IsNewRecord(record: boolean) {
    this._IsNewRecord = record;
  }

  public get DocType() {
    return this._DocType;
  }

  public set DocType(docType: DocType) {
    this._DocType = docType;
  }

  public get DocDate() {
    return this._DocDate;
  }

  public set DocDate(date: Date) {
    this._DocDate = date;
  }

  public get DocHTMLFileMediaId() {
    return this._DocHTMLFileMediaId;
  }

  public set DocHTMLFileMediaId(mediaId: string) {
    this._DocHTMLFileMediaId = mediaId;
  }

  public get DocPDFFileMediaId() {
    return this._DocPDFFileMediaId;
  }

  public set DocPDFFileMediaId(mediaId: string) {
    this._DocPDFFileMediaId = mediaId;
  }

  public get Amount() {
    return this._Amount;
  }

  private set Amount(amount: number) {
    this._Amount = amount;
  }

  public get Status() {
    return this._Status;
  }

  private set Status(status: DocumentStatus) {
    this._Status = status;
  }

  /* AccountSystemEntity Implementation */
  get RepositoryBase() {
    return Document._RepositoryBase;
  }

  /* ObjectBase Implementation */
  get ObjectId() {
    return this.DocNo;
  }

  get ObjectName() {
    return this.DocNo;
  }

  get TableName() {
    return 'finance_Document';
  }

  constructor(dbTransaction?: any, docNo?: string) {
    super();
    if (dbTransaction) {
      this._DbTransaction = dbTransaction;
    }
    if (docNo) {
      this.DocNo = docNo;
      this.RepositoryBase.findOne({
        where: {
          DocNo: this.DocNo,
        },
        transaction: dbTransaction,
      })
        .then((documentData) => {
          if (documentData) {
            this.IsNewRecord = false;
            this.DocType = documentData.DocType;
            this.DocDate = documentData.DocDate;
            this.CompanyId = documentData.CompanyId;
            this.Currency = documentData.Currency;
            this.Amount = documentData.Amount;
            this.Description = documentData.Description;
            this.Status = documentData.Status;
            this.IssuedById = documentData.IssuedById;
            this.IssuedToId = documentData.IssuedToId;
            this.IssuedToType = documentData.IssuedToType;
            this.RelatedObjectId = documentData.RelatedObjectId;
            this.RelatedObjectType = documentData.RelatedObjectType;
            this.CreatedById = documentData.CreatedById;
            this.CreatedAt = documentData.CreatedAt;
            this.UpdatedById = documentData.UpdatedById;
            this.UpdatedAt = documentData.UpdatedAt;
            this.DocPDFFileMediaId = documentData.DocPDFFileMediaId;
            this.DocHTMLFileMediaId = documentData.DocHTMLFileMediaId;
            this.AccSystemRefId = documentData.AccSystemRefId;
            this.PostedToAccSystemYN = documentData.PostedToAccSystemYN;
            this.PostedById = documentData.PostedById;
            this.PostedDateTime = documentData.PostedDateTime;
            this.UseAccSystemDocYN = documentData.UseAccSystemDocYN;
            this.IsNewRecord = false;
          } else {
            const notFoundError = new RecordNotFoundError('No Record Found.');
            throw notFoundError;
          }
        })
        .catch((err) => {
          // tslint:disable-next-line:no-console
          console.log('Document Class constructor err: ', err);
        });
    }
  }

  public static async initDocument(dbTransaction: any, docNo: string) {
    const document = new Document(dbTransaction);
    const documentData = await this._RepositoryBase.findOne({
      where: {
        DocNo: docNo,
      },
    });
    if (documentData) {
      document.IsNewRecord = false;
      document.DocType = documentData.DocType;
      document.DocDate = documentData.DocDate;
      document.CompanyId = documentData.CompanyId;
      document.Currency = documentData.Currency;
      document.Amount = documentData.Amount;
      document.Description = documentData.Description;
      document.Status = documentData.Status;
      document.IssuedById = documentData.IssuedById;
      document.IssuedToId = documentData.IssuedToId;
      document.IssuedToType = documentData.IssuedToType;
      document.RelatedObjectId = documentData.RelatedObjectId;
      document.RelatedObjectType = documentData.RelatedObjectType;
      document.CreatedById = documentData.CreatedById;
      document.CreatedAt = documentData.CreatedAt;
      document.UpdatedById = documentData.UpdatedById;
      document.UpdatedAt = documentData.UpdatedAt;
      document.DocPDFFileMediaId = documentData.DocPDFFileMediaId;
      document.DocHTMLFileMediaId = documentData.DocHTMLFileMediaId;
      document.AccSystemRefId = documentData.AccSystemRefId;
      document.PostedToAccSystemYN = documentData.PostedToAccSystemYN;
      document.PostedById = documentData.PostedById;
      document.PostedDateTime = documentData.PostedDateTime;
      document.UseAccSystemDocYN = documentData.UseAccSystemDocYN;
      document.IsNewRecord = false;
    } else {
      const notFoundError = new RecordNotFoundError('No Record Found.');
      throw notFoundError;
    }
    return document;
  }

  /**
   * Recalculate the Amount of the Document based on all the amount of the Document Items
   *
   * @returns {void}
   */
  async reCalculateAmount(): Promise<void> {
    this.Amount = 0;
    for (const documentItem of this._DocumentItems) {
      this.Amount += documentItem.Amount;
    }
  }

  get DocumentItems(): Promise<DocumentItem[]> {
    return new Promise((resolve, reject) => {
      if (!this.IsNewRecord) {
        Document._DocumentItemRepository
          .findAll({
            where: {
              DocNo: this.DocNo,
            },
            transaction: this._DbTransaction,
          })
          .then((documentItems) => {
            const documentItemObjects = documentItems.map((documentItem) => {
              const di = new DocumentItem(this._DbTransaction, this);
              di.init({ ...documentItem.get({ plain: true }) });
            });
            return Promise.all(documentItemObjects);
          })
          .then((documentItemObjects) => {
            this._DocumentItems = documentItemObjects;
            resolve(this._DocumentItems);
          })
          .catch((err) => {
            reject(err);
          });
      } else {
        resolve(this._DocumentItems);
      }
    });
  }

  /**
   * Returns HTML template file for invoice generation
   *
   * @see generateInvoice - for implementation
   *
   * @returns {Express.Multer.File} - HTML File
   */
  public static async DEFAULT_INVOICE_TEMPLATE_HTML(): Promise<Express.Multer.File> {
    /* getting invoice template in HTML format */
    const templateLocation = path.resolve('./invoice-template/index.html');
    const readFileContent = util.promisify(fs.readFile);
    const htmlBuffer = await readFileContent(templateLocation);

    const htmlStream: Express.Multer.File = {
      buffer: htmlBuffer,
      fieldname: 'FileStream',
      originalname: 'index.html',
      mimetype: 'text/html',
      size: htmlBuffer.length,
      encoding: '8bit',
      filename: 'sample-invoice.html',
      destination: '',
      path: '',
      stream: null,
    };

    return htmlStream;
  }

  /**
   * Returns PDF template file for invoice generation
   *
   * @remarks - This function is converting generated HTML into PDF
   *
   * @see generateInvoice - for implementation
   *
   * @returns {Express.Multer.File} - PDF File
   */
  public static async DEFAULT_INVOICE_TEMPLATE_PDF(): Promise<Express.Multer.File> {
    /* prepping HTML conversion by getting its URL and String */

    /* todo: needs to generate HTML first before converting to PDF */
    const templateLocation = path.resolve('./invoice-template/index.html');
    const readFileContent = util.promisify(fs.readFile);
    const htmlString = await readFileContent(templateLocation, 'utf8');

    /* converting HTML to PDF */
    const browser = await puppeteer.launch();
    const page = await browser.newPage();

    if (htmlString) {
      await page.setContent(htmlString);
    } else if (templateLocation) {
      await page.goto(templateLocation, {
        waitUntil: 'networkidle2',
      });
    } else {
      await browser.close();
    }

    const pdfBuffer = await page.pdf({
      format: 'a4',
      printBackground: true,
      margin: {
        left: '0px',
        top: '0px',
        right: '0px',
        bottom: '0px',
      },
    });

    const pdfStream: Express.Multer.File = {
      buffer: pdfBuffer,
      fieldname: 'FileStream',
      originalname: 'sample-invoice.pdf',
      mimetype: 'application/pdf',
      size: pdfBuffer.length,
      encoding: '8bit',
      filename: 'sample-invoice.pdf',
      destination: '',
      path: '',
      stream: null,
    };

    return pdfStream;
  }

  /**
   * Returns the media representing the document in HTML format
   *
   * @returns {MediasModel} - Invoice Media in HTML Format
   */
  public get DocumentFileInHTML(): Promise<MediasModel> {
    return (async () => {
      const { HTMLMedia } = await this.generateInvoice();

      return HTMLMedia;
    })();
  }

  /**
   * Returns the media representing the document in PDF format
   *
   * @returns {MediasModel} - Invoice Media in PDF Format
   */
  public get DocumentFileInPDF(): Promise<MediasModel> {
    return (async () => {
      const { PDFMedia } = await this.generateInvoice();

      return PDFMedia;
    })();
  }

  init(params: IDocumentAttr) {
    this.DocNo = params.DocNo;
    this.DocType = params.DocType;
    this.DocDate = params.DocDate;
    this.CompanyId = params.CompanyId;
    this.Currency = params.Currency;
    this.Amount = params.Amount;
    this.Description = params.Description;
    this.Status = params.Status;
    this.IssuedById = params.IssuedById;
    this.IssuedToId = params.IssuedToId;
    this.IssuedToType = params.IssuedToType;
    this.RelatedObjectId = params.RelatedObjectId;
    this.RelatedObjectType = params.RelatedObjectType;
    this.CreatedById = params.CreatedById;
    this.CreatedAt = params.CreatedAt;
    this.UpdatedById = params.UpdatedById;
    this.UpdatedAt = params.UpdatedAt;
    this.DocPDFFileMediaId = params.DocPDFFileMediaId;
    this.DocHTMLFileMediaId = params.DocHTMLFileMediaId;
    this.AccSystemRefId = params.AccSystemRefId;
    this.PostedToAccSystemYN = params.PostedToAccSystemYN;
    this.PostedById = params.PostedById;
    this.PostedDateTime = params.PostedDateTime;
    this.UseAccSystemDocYN = params.UseAccSystemDocYN;
  }

  /**
   * Returns generated HTML & PDF Media
   *
   * @remarks - This function also create each media element in common_Media table
   *
   * @see DocumentFileInHTML - for implementation
   * @see DocumentFileInPDF - for implementation
   *
   * @see {@link FinanceCompany.issueInvoice} - for implementation
   *
   * @param userId - userId that trigger this function
   *
   * @returns {MediasModel} - HTML & PDF Media
   */
  async generateInvoice(
    userId?: string,
    customer?: FinanceCustomerBase,
  ): Promise<{
    HTMLMedia: MediasModel;
    PDFMedia: MediasModel;
  }> {
    const media: Medias = new Medias(
      Document._MediaRepository,
      this.mediaCommonService,
    );

    /* todo: needs to manipulate HTML DOM before uploading to media*/
    console.log(customer || 'no customer provided', '<< customer');

    /* insert FileStream on htmlPayload  */
    const htmlPayload: InternalMediaDto = {
      ObjectId: this.DocNo,
      ObjectType: this.DocType,
      Type: MediaType.Document,
      FileName: `${this.DocType}-${this.DocNo}`,
      FileStream: await Document.DEFAULT_INVOICE_TEMPLATE_HTML(),
      FileExtension: 'html',
      Title: `${this.DocType}-${this.DocNo}.html`,
      Description: `HTML ${this.DocType}`,
      IsEncryptedYN: 'N',
    };

    /* create HTML media of the generated HTML */
    const HTMLInvoiceMedia = await media.postInternal(
      await Document.DEFAULT_INVOICE_TEMPLATE_HTML(),
      htmlPayload,
      userId ?? 'System',
    );

    /* insert FileStream on pdfPayload  */
    const pdfPayload: InternalMediaDto = {
      ObjectId: this.DocNo,
      ObjectType: this.DocType,
      Type: MediaType.Document,
      FileName: `${this.DocType}-${this.DocNo}`,
      FileStream: await Document.DEFAULT_INVOICE_TEMPLATE_PDF(),
      FileExtension: 'pdf',
      Title: `${this.DocType}-${this.DocNo}.pdf`,
      Description: `PDF ${this.DocType}`,
      IsEncryptedYN: 'N',
    };

    /* create PDF media from the converted HTML */
    const PDFInvoiceMedia = await media.postInternal(
      await Document.DEFAULT_INVOICE_TEMPLATE_PDF(),
      pdfPayload,
      userId ?? 'System',
    );

    try {
      const document = await this.RepositoryBase.findOne({
        where: {
          DocNo: this.DocNo,
        },
        transaction: this._DbTransaction,
      });

      /* storing HTML & PDF MediaId to finance_Document */
      document.DocHTMLFileMediaId = HTMLInvoiceMedia.MediaId;
      document.DocPDFFileMediaId = PDFInvoiceMedia.MediaId;
      await document.save({ transaction: this._DbTransaction });
    } catch (err) {
      // tslint:disable-next-line:no-console
      console.log(err);
    }

    return {
      HTMLMedia: HTMLInvoiceMedia,
      PDFMedia: PDFInvoiceMedia,
    };
  }

  async generateCreditNote(userId?: string, customer?: FinanceCustomerBase) {
    console.log(userId, customer, '<< generateCreditNote');
  }

  /**
   * Create a new Document Item which is automatically added to the DocumentItems list
   * @param documentItem
   *
   * @returns {DocumentItem}
   */
  async newDocumentItem(documentItem?: DocumentItem): Promise<DocumentItem> {
    let di: DocumentItem;
    if (documentItem) {
      di = documentItem;
    } else {
      di = new DocumentItem(this._DbTransaction, this);
    }
    di.DocNo = this.DocNo;
    this._DocumentItems.push(di);
    return documentItem;
  }

  /**
   * this is temporary method to get the latest previous document
   *
   * @see SimpleLoans.BillLatePaymentCharges - for implementation
   *
   * @param docType - The document type
   *
   * @returns {DocumentModel}
   */
  async getPreviousDocument(docType: DocType): Promise<DocumentModel> {
    return await this.RepositoryBase.findOne({
      where: {
        DocType: docType,
      },
      order: [['CreatedAt', 'DESC']],
    });
  }
}
