import axios from 'axios';
import * as cuid from 'cuid';
import {
  HashTable,
  LoginUserBase,
  ObjectBase,
  RecordNotFoundError,
} from '@tomei/general';
import Account from '../account/account';
import JournalEntry from '../journal-entry/journal-entry';
import FinanceCustomerBase from '../customer/customer';
import Document from '../document/document';
import { IAccountSystem } from '../interfaces';
import { FinanceCompanyRepository } from './finance-company.repository';
import { FinanceCustomerRepository } from '../customer/finance-customer.repository';
import { LedgerTransactionRepository } from '../ledger-transaction/ledger-transaction.repository';
import {
  DocType,
  PaymentStatus,
  PaymentType,
  TransactionTypeOptions,
} from '../enum';
import PaymentMethodType from '../payment-method-type/payment-method-type';
import { PaymentRepository } from '../payment/payment.repository';
import { PaymentItemRepository } from '../payment-item/payment-item.repository';
import Payment from '../payment/payment';
import { DocumentRepository } from '../document/document.repository';
import { DocumentItemRepository } from '../document/document-item.repository';
import { PaymentMethodRepository } from '../payment-method/payment-method.repository';
import { PaymentMethodTypeRepository } from '../payment-method-type/payment-method-type.repository';
import PaymentMethod from '../payment-method/payment-method';
import { AccountRepository } from '../account/account.repository';
import { PaymentPaidWithRepository } from '../payment-paid-with/payment-paid-with.repository';

// eslint-disable-next-line @typescript-eslint/no-var-requires
const getConfig = require('../config');
// eslint-disable-next-line @typescript-eslint/no-var-requires
const config = getConfig();

export default class FinanceCompany extends ObjectBase {
  private _CompanyId = 'New';
  private _CompSystemCode = '';
  private _CompSystemRefId = '';
  private _AccSystemCode = '';
  private static _htFinanceCompanyIds = new HashTable();
  private static _htFinanceCompanies = new HashTable();
  private static _financeCompanyRepository = new FinanceCompanyRepository();
  private static _PaymentRepository = new PaymentRepository();
  private static _PaymentItemRepository = new PaymentItemRepository();
  private static _PaymentPaidWithRepository = new PaymentPaidWithRepository();
  private static _PaymentMethodRepository = new PaymentMethodRepository();
  private static _PaymentMethodTypeRepository =
    new PaymentMethodTypeRepository();
  private static _DocumentRepository = new DocumentRepository();
  private static _DocumentItemRepository = new DocumentItemRepository();
  private static _FinanceCustomerRepository = new FinanceCustomerRepository();
  private static _LedgerTransactionRepository =
    new LedgerTransactionRepository();

  private static _AccountRepository = new AccountRepository();

  private _AccountingSystem: IAccountSystem;

  private _DbTransaction: any;
  private _PaymentMethods = [];

  get CompSystemCode(): string {
    return this._CompSystemCode;
  }

  private set CompSystemCode(code: string) {
    this._CompSystemCode = code;
  }

  get CompSystemRefId() {
    return this._CompSystemRefId;
  }

  set CompSystemRefId(id: string) {
    this._CompSystemRefId = id;
  }

  get AccSystemCode() {
    return this._AccSystemCode;
  }

  private set AccSystemCode(code: string) {
    this._AccSystemCode = code;
  }

  get CompanyId() {
    return this._CompanyId;
  }

  get ObjectId() {
    return this._CompanyId;
  }

  private set ObjectId(id: string) {
    this._CompanyId = id;
  }

  get ObjectName() {
    return `${this.CompSystemCode}-${this.CompSystemRefId}-${this.AccSystemCode}`;
  }

  get TableName() {
    return 'finance_Company';
  }

  get AccountingSystem(): IAccountSystem {
    return this._AccountingSystem;
  }

  set AccountingSystem(system: IAccountSystem) {
    this._AccountingSystem = system;
  }

  constructor(
    compSystemCode: string,
    compSystemRefId: string,
    accSystemCode: string,
  ) {
    super();
    this.CompSystemCode = compSystemCode;
    this.CompSystemRefId = compSystemRefId;
    this.AccSystemCode = accSystemCode;
  }

  static async getFinanceCompanyId(
    compSystemCode: string,
    compSystemRefId: string,
    accSystemCode: string,
  ): Promise<string> {
    let sCompanyId = '';
    /*Assemble the hashtable key*/
    const sKey = `${compSystemCode}-${compSystemRefId}-${accSystemCode}`;
    /*Check if the FinanceCompany has previously being loaded*/
    if (!FinanceCompany._htFinanceCompanyIds.get(sKey)) {
      /*Instantiate a new FinanceCompany*/
      const financeCompany = new FinanceCompany(
        compSystemCode,
        compSystemRefId,
        accSystemCode,
      );

      /*Retrieve the finance company from finance_Company table using compSystemCode,
       * CompSystemRefId and accSystemCode */
      const company = await FinanceCompany._financeCompanyRepository.findOne({
        where: {
          CompSystemCode: compSystemCode,
          CompSystemRefId: compSystemRefId,
          AccSystemCode: accSystemCode,
        },
      });

      /*Retrieve and store the companyId from the result*/
      financeCompany.ObjectId = company.CompanyId;
      sCompanyId = financeCompany.ObjectId;

      /*Add the details into the hashtable*/
      FinanceCompany._htFinanceCompanyIds.add(sKey, financeCompany.ObjectId);
      FinanceCompany._htFinanceCompanies.add(sCompanyId, financeCompany);
    }

    if (typeof FinanceCompany._htFinanceCompanyIds.get(sKey) === 'string') {
      sCompanyId = FinanceCompany._htFinanceCompanyIds.get(sKey);
    }

    return sCompanyId;
  }

  static async getFinanceCompany(companyId: string): Promise<FinanceCompany> {
    /*Check if the finance company is previously be loaded*/
    if (!FinanceCompany._htFinanceCompanies.get(companyId)) {
      /*Retrieve the finance company from finance_Company table using compSystemCode,
       * CompSystemRefId and accSystemCode */
      const company = await FinanceCompany._financeCompanyRepository.findOne({
        where: { CompanyId: companyId },
      });

      if (!company) {
        throw Error('No finance company found. Please create first.');
      }

      /*Using the result returned, instantiate a new FinanceCompany*/
      const compSystemCode = company.CompSystemCode;
      const compSystemRefId = company.CompSystemRefId;
      const accSystemCode = company.AccSystemCode;

      const financeCompany = new FinanceCompany(
        compSystemCode,
        compSystemRefId,
        accSystemCode,
      );

      /*Retrieve and store the CompanyId from the result*/
      financeCompany.ObjectId = company.CompanyId;
      financeCompany._CompanyId = company.CompanyId;

      /*Add the details into the hashtable*/
      const sKey = `${compSystemCode}-${compSystemRefId}-${accSystemCode}`;
      FinanceCompany._htFinanceCompanyIds.add(sKey, financeCompany.ObjectId);
      FinanceCompany._htFinanceCompanies.add(
        financeCompany.ObjectId,
        financeCompany,
      );
    }
    // tslint:disable-next-line:no-console
    console.log('return from hash table');
    return FinanceCompany._htFinanceCompanies.get(companyId);
  }

  static async createFinanceCompany(
    dbTransaction: any,
    loginUser: LoginUserBase,
    companyId: string,
    compSystemCode: string,
    compSystemRefId: string,
    accSystemCode: string,
  ) {
    /*Instantiate a new FinanceCompany*/
    const financeCompany = new FinanceCompany(
      compSystemCode,
      compSystemRefId,
      accSystemCode,
    );

    /*Validating if the finance company already exists in finance_Company*/
    const company = await FinanceCompany._financeCompanyRepository.findOne({
      where: {
        CompSystemCode: compSystemCode,
        CompSystemRefId: compSystemRefId,
        AccSystemCode: accSystemCode,
      },
    });

    if (company) {
      throw Error(
        'There is already another Finance Company with the compSystemCode, CompSystemRefId and accSystemCode specified.',
      );
    }

    /*Generating the companyId*/
    financeCompany.ObjectId = companyId;

    /*Save the FinanceCompany to the finance_Company table*/
    await FinanceCompany._financeCompanyRepository.create(
      {
        CompanyId: financeCompany.CompanyId,
        CompSystemCode: financeCompany.CompSystemCode,
        CompSystemRefId: financeCompany.CompSystemRefId,
        AccSystemCode: financeCompany.AccSystemCode,
      },
      {
        transaction: dbTransaction,
      },
    );

    /*Add the details to hashtable*/
    const sKey = `${compSystemCode}-${compSystemRefId}-${accSystemCode}`;
    FinanceCompany._htFinanceCompanyIds.add(sKey, financeCompany.ObjectId);
    FinanceCompany._htFinanceCompanies.add(
      financeCompany.ObjectId,
      financeCompany,
    );

    // tslint:disable-next-line:no-console
    console.log('return from hash table');
    return FinanceCompany._htFinanceCompanies.get(companyId);
  }

  async createCustomer(
    dbTransaction: any,
    custSystemCode: string,
    custSystemRefId: string,
    customer: FinanceCustomerBase,
  ) {
    try {
      if (!custSystemCode || !custSystemRefId) {
        throw new Error(
          'CustSystemCode and CustomerRefId are required fields.',
        );
      }

      const financeCustomerData =
        await FinanceCompany._FinanceCustomerRepository.findOne({
          where: {
            CompanyId: this._CompanyId,
            CustSystemCode: custSystemCode,
            CustSystemRefId: custSystemRefId,
          },
        });

      if (financeCustomerData) {
        throw new Error('Customer already created previously.');
      }

      // Retrieve the type of accounting system, API Key, and API secret based on SysCode from the Config.ts
      const customerId = await this.AccountingSystem.createCustomer({
        customer,
      });

      customer.CompanyId = this._CompanyId;

      const newCustomer = await customer.save(
        customerId,
        custSystemCode,
        custSystemRefId,
        dbTransaction,
      );

      const payload = {
        Action: 'Create',
        Activity: 'Finance Customer Created',
        Description: `Customer (ID: ${newCustomer.CustomerId}) has been created for ${this.AccSystemCode}`,
        EntityType: 'finance_Customer',
        EntityValueBefore: JSON.stringify({}),
        EntityValueAfter: JSON.stringify(newCustomer),
        PerformedById: 'test',
        PerformedAt: new Date(),
        EntityId: newCustomer.CustomerId,
      };

      await axios.post(
        `${process.env.COMMON_API_URL}/activity-histories`,
        payload,
      );

      return customer;
    } catch (error) {
      throw error;
    }
  }

  async postJournal(dbTransaction: any, journalEntry: JournalEntry) {
    const debitTransactions = await journalEntry.DebitTransactions;
    const creditTransactions = await journalEntry.CreditTransactions;
    try {
      if (creditTransactions.length < 1 || debitTransactions?.length < 1) {
        throw new Error(
          'There should be at least 1 debit ledger transaction and 1 credit ledger transaction in the journal entry',
        );
      }

      const totalCreditAmount = creditTransactions.reduce(
        (accumulator, currentValue) => accumulator + currentValue.CreditAmount,
        0,
      );
      const totalDebitAmount = debitTransactions.reduce(
        (accumulator, currentValue) => accumulator + currentValue.DebitAmount,
        0,
      );

      if (totalCreditAmount !== totalDebitAmount) {
        throw new Error(
          'Credit ledger transaction and debit ledger transaction should the same amount',
        );
      }

      const newJournalEntry = await journalEntry.save('test', dbTransaction);

      const allLedgerTransactions =
        await FinanceCompany._LedgerTransactionRepository.findAll({});

      for (const ledgerTransaction of allLedgerTransactions) {
        ledgerTransaction.JournalEntryId = newJournalEntry.JournalEntryId;
        await ledgerTransaction.save();
      }

      await this.AccountingSystem.postJournalEntry(newJournalEntry);

      const payload = {
        Action: 'Create',
        Activity: 'Post Journal Entry',
        Description: `Journal Entry (ID: ${newJournalEntry.JournalEntryId}) has been created`,
        EntityType: 'JournalEntry',
        EntityValueBefore: JSON.stringify({}),
        EntityValueAfter: JSON.stringify(newJournalEntry),
        PerformedById: 'test',
        PerformedAt: new Date(),
        EntityId: newJournalEntry.JournalEntryId,
      };

      await axios.post(
        `${process.env.COMMON_API_URL}/activity-histories`,
        payload,
      );
    } catch (error) {
      throw error;
    }
  }

  async createAccount(dbTransaction: any, account: Account) {
    try {
      account.validateAccountValue();

      let createAccountPayload: any = {
        Name: account.Name,
        AcctNum: account.AccountNo,
        AccountType: account.AccountType,
        AccountSubType: account.AccountSubtype,
      };

      if (account.isParentAccountExists()) {
        createAccountPayload = {
          ...createAccountPayload,
          CurrencyRef: 'MYR',
          ParentRef: account.ParentAccountNo,
          SubAccount: true,
        };
      }

      const accSystemAccountId = await this.AccountingSystem.createAccount(
        createAccountPayload,
      );

      const newAccount = await account.save(
        accSystemAccountId,
        'test',
        dbTransaction,
      );

      const payload = {
        Action: 'Create',
        Activity: 'Account Created',
        Description: `Account (ID: ${newAccount.AccountNo}) has been created`,
        EntityType: 'Account',
        EntityValueBefore: JSON.stringify({}),
        EntityValueAfter: JSON.stringify(newAccount),
        PerformedById: 'test',
        PerformedAt: new Date(),
        EntityId: newAccount.AccountNo,
      };

      await axios.post(
        `${process.env.COMMON_API_URL}/activity-histories`,
        payload,
      );

      return account;
    } catch (error) {
      throw error;
    }
  }

  /**
   * Issue an invoice
   *
   * @param dbTransaction - The database transaction to be used
   * @param loginUser - The user issuing the invoice
   * @param invoice - The document containing the invoice details
   * @param customer - The customer to be issued the invoice
   * @param dtAccountNo - The account number of the Account Receivable (AR) account to debit. If not provided, will debit the customer's default AR account
   *
   * @returns {Document} - Document object representing the full details of the invoice issued
   */
  async issueInvoice(
    /*todo: loginUser & customer is NOT supposed to be optional */
    dbTransaction: any,
    invoice: Document,
    loginUser?: LoginUserBase,
    customer?: FinanceCustomerBase,
    dtAccountNo?: string,
  ): Promise<Document> {
    try {
      /*Check if the invoice number already exists (should be unique)*/
      const duplicateInvoice = await FinanceCompany._DocumentRepository.findOne(
        {
          where: {
            DocNo: invoice.DocNo,
          },
          transaction: dbTransaction,
        },
      );

      if (duplicateInvoice) {
        throw new Error('Invoice number already exists');
      }

      const documentItems = await invoice.DocumentItems;

      /*Check if the document has at least 1 document item*/
      if (!documentItems.length) {
        throw new Error('Document must have at least 1 document item');
      }

      /*Check if each document item has CtAccountNo provided*/
      for (const invoiceItem of documentItems) {
        if (!invoiceItem.CtAccountNo) {
          throw new Error(
            'Each document item should have CtAccountNo provided',
          );
        }
      }

      /*Set up the document type*/
      invoice.DocType = DocType.INVOICE;
      // invoice.DocNo = cuid();

      /*Saving the document and document items to the database*/
      await FinanceCompany._DocumentRepository.create(
        {
          DocNo: invoice.DocNo,
          DocType: invoice.DocType,
          DocDate: invoice.DocDate,
          CompanyId: invoice.CompanyId,
          Currency: invoice.Currency,
          Amount: invoice.Amount,
          Description: invoice.Description,
          Status: invoice.Status,
          IssuedById: invoice.IssuedById,
          IssuedToId: invoice.IssuedToId,
          IssuedToType: invoice.IssuedToType,
          RelatedObjectId: invoice.RelatedObjectId,
          RelatedObjectType: invoice.RelatedObjectType,
          CreatedById: invoice.CreatedById,
          CreatedAt: new Date(),
          UpdatedById: invoice.UpdatedById,
          UpdatedAt: new Date(),
          DocPDFFileMediaId: invoice.DocPDFFileMediaId,
          DocHTMLFileMediaId: invoice.DocHTMLFileMediaId,
          AccSystemRefId: invoice.AccSystemRefId,
          PostedToAccSystemYN: invoice.PostedToAccSystemYN,
          PostedById: invoice.PostedById,
          PostedDateTime: new Date(),
          UseAccSystemDocYN: invoice.UseAccSystemDocYN,
        },
        {
          transaction: dbTransaction,
        },
      );

      for (const documentItem of documentItems) {
        await FinanceCompany._DocumentItemRepository.create(
          {
            DocumentItemId: cuid(),
            DocNo: invoice.DocNo,
            Name: documentItem.Name,
            NameBM: documentItem.NameBM,
            Description: documentItem.Description,
            ItemId: documentItem.ItemId,
            ItemType: documentItem.ItemType,
            ItemSKU: documentItem.ItemSKU,
            ItemSerialNo: documentItem.ItemSerialNo,
            Currency: documentItem.Currency,
            UnitPrice: documentItem.UnitPrice,
            Quantity: documentItem.Quantity,
            QuantityUOM: documentItem.QuantityUOM,
            Amount: documentItem.Amount,
            TaxCode: documentItem.TaxCode,
            TaxAmount: documentItem.TaxAmount,
            TaxRate: documentItem.TaxRate,
            TaxInclusiveYN: documentItem.TaxInclusiveYN,
            DtAccountNo: documentItem.DtAccountNo
              ? documentItem.DtAccountNo
              : null,
            CtAccountNo: documentItem.CtAccountNo
              ? documentItem.CtAccountNo
              : null,
          },
          {
            transaction: dbTransaction,
          },
        );
      }

      /*Generating the invoice*/
      if (invoice.UseAccSystemDocYN === 'Y') {
        /*todo: Posting to accounting system to generate invoice*/
        await this.AccountingSystem.createInvoice(invoice);
      } else {
        /*todo: check config file to see which invoice template is to be used for specific project*/

        /*Generating invoice based on template*/
        invoice.generateInvoice(invoice.IssuedById, customer);
      }

      const transactionDate = new Date();
      const htCreditAccountAmount = new HashTable();
      const htCreditAccountCurrency = new HashTable();

      documentItems.forEach((invoiceItem) => {
        if (!htCreditAccountAmount.exists(invoiceItem.CtAccountNo)) {
          //add the credit account to the hash table
          htCreditAccountAmount.add(
            invoiceItem.CtAccountNo,
            invoiceItem.Amount,
          );

          htCreditAccountCurrency.add(
            invoiceItem.CtAccountNo,
            invoiceItem.Currency,
          );
        } else {
          //update the credit account amount
          const d = htCreditAccountAmount.get(invoiceItem.CtAccountNo);
          htCreditAccountAmount.add(
            invoiceItem.CtAccountNo,
            d + invoiceItem.Amount,
          );
        }
      });

      const savedItems = htCreditAccountAmount.list();

      for (let i = 0; i < savedItems.length; i++) {
        const journalEntry = new JournalEntry(dbTransaction);
        //Temporary fix to successfully save journal entry in PostJournal
        journalEntry.init({
          CompanyId: this.CompanyId,
          Name: 'Issue Invoice ' + invoice.DocNo,
        });
        const account = await Account.initAccount(
          dbTransaction,
          savedItems[i].value[0],
        );
        const creditAmount = savedItems[i].value[1];
        const currency = htCreditAccountCurrency.get(savedItems[i].value[0]);

        const dt = await journalEntry.newLedgerTransaction(
          TransactionTypeOptions.DEBIT,
        );

        if (dtAccountNo) {
          /*Transacting using AR account provided*/
          dt.AccountNo = dtAccountNo;
        } else {
          /*Transacting based on default customer AR account*/
          dt.AccountNo = (await customer.AccountReceivable).AccountNo;
        }

        dt.Currency = currency ? currency : 'MYR';
        dt.DebitAmount = creditAmount ? creditAmount : 0.0;
        dt.Date = transactionDate;
        dt.Description = account.Name;
        dt.Name = account.Name;
        dt.RelatedDocNo = invoice.DocNo;

        const ct = await journalEntry.newLedgerTransaction(
          TransactionTypeOptions.CREDIT,
        );
        ct.AccountNo = savedItems[i].key;
        ct.Currency = currency ? currency : 'MYR';
        ct.CreditAmount = creditAmount ? creditAmount : 0.0;
        ct.Date = transactionDate;
        ct.Description = account.Name;
        ct.Name = account.Name;
        ct.RelatedDocNo = invoice.DocNo;

        await this.postJournal(dbTransaction, journalEntry);
      }

      return invoice;
    } catch (err) {
      // tslint:disable-next-line:no-console
      console.log('Issue invoice err: ', err);
      throw err;
    }
  }

  /**
   * Issue a debit note
   *
   * @param dbTransaction - The database transaction to be used
   * @param loginUser - The user issuing the invoice
   * @param invoice - The document containing the invoice details
   * @param customer - The customer to be issued the invoice
   * @param dtAccountNo - The account number of the Account Receivable (AR) account to debit. If not provided, will debit the customer's default AR account
   *
   * @returns {Document} - Document object representing the full details of the invoice issued
   */
  async issueDebitNote(
    dbTransaction: any,
    loginUser: LoginUserBase,
    invoice: Document,
    customer: FinanceCustomerBase,
    dtAccountNo?: string,
  ): Promise<Document> {
    try {
      /*Check if the invoice number already exists (should be unique)*/
      const duplicateInvoice = await FinanceCompany._DocumentRepository.findOne(
        {
          where: {
            DocNo: invoice.DocNo,
          },
          transaction: dbTransaction,
        },
      );

      if (duplicateInvoice) {
        throw new Error('Invoice number already exists');
      }

      const documentItems = await invoice.DocumentItems;

      /*Check if the document has at least 1 document item*/
      if (!documentItems.length) {
        throw new Error('Document must have at least 1 document item');
      }

      /*Check if each document item has CtAccountNo provided*/
      for (const invoiceItem of documentItems) {
        if (!invoiceItem.CtAccountNo) {
          throw new Error(
            'Each document item should have CtAccountNo provided',
          );
        }
      }

      /*Set up the document type*/
      invoice.DocType = DocType.DEBIT_NOTE;
      // invoice.DocNo = cuid();

      /*Saving the document and document items to the database*/
      await FinanceCompany._DocumentRepository.create(
        {
          DocNo: invoice.DocNo,
          DocType: invoice.DocType,
          DocDate: invoice.DocDate,
          CompanyId: invoice.CompanyId,
          Currency: invoice.Currency,
          Amount: invoice.Amount,
          Description: invoice.Description,
          Status: invoice.Status,
          IssuedById: loginUser.ObjectId,
          IssuedToId: customer.ObjectId,
          IssuedToType: invoice.IssuedToType,
          RelatedObjectId: invoice.RelatedObjectId,
          RelatedObjectType: invoice.RelatedObjectType,
          CreatedById: loginUser.ObjectId,
          CreatedAt: new Date(),
          UpdatedById: loginUser.ObjectId,
          UpdatedAt: new Date(),
          DocPDFFileMediaId: invoice.DocPDFFileMediaId,
          DocHTMLFileMediaId: invoice.DocHTMLFileMediaId,
          AccSystemRefId: invoice.AccSystemRefId,
          PostedToAccSystemYN: invoice.PostedToAccSystemYN,
          PostedById: invoice.PostedById,
          PostedDateTime: new Date(),
          UseAccSystemDocYN: invoice.UseAccSystemDocYN,
        },
        {
          transaction: dbTransaction,
        },
      );

      documentItems.forEach(async (documentItem) => {
        await FinanceCompany._DocumentItemRepository.create(
          {
            DocumentItemId: cuid(),
            DocNo: documentItem.DocNo,
            Name: documentItem.Name,
            NameBM: documentItem.NameBM,
            Description: documentItem.Description,
            ItemId: documentItem.ItemId,
            ItemType: documentItem.ItemType,
            ItemSKU: documentItem.ItemSKU,
            ItemSerialNo: documentItem.ItemSerialNo,
            Currency: documentItem.Currency,
            UnitPrice: documentItem.UnitPrice,
            Quantity: documentItem.Quantity,
            QuantityUOM: documentItem.QuantityUOM,
            Amount: documentItem.Amount,
            TaxCode: documentItem.TaxCode,
            TaxAmount: documentItem.TaxAmount,
            TaxRate: documentItem.TaxRate,
            TaxInclusiveYN: documentItem.TaxInclusiveYN,
            DtAccountNo: documentItem.DtAccountNo,
            CtAccountNo: documentItem.CtAccountNo,
          },
          {
            transaction: dbTransaction,
          },
        );
      });

      /*Generating the invoice*/
      if (invoice.UseAccSystemDocYN === 'Y') {
        /*todo: Posting to accounting system to generate invoice*/
        await this.AccountingSystem.createInvoice(invoice);
      } else {
        /*todo: check config file to see which invoice template is to be used for specific project*/

        /*Generating invoice based on template*/
        invoice.generateInvoice(loginUser.IDNo, customer);
      }

      const transactionDate = new Date();
      const htCreditAccountAmount = new HashTable();
      const htCreditAccountCurrency = new HashTable();

      documentItems.forEach((invoiceItem) => {
        if (!htCreditAccountAmount.exists(invoiceItem.CtAccountNo)) {
          //add the credit account to the hash table
          htCreditAccountAmount.add(
            invoiceItem.CtAccountNo,
            invoiceItem.Amount,
          );

          htCreditAccountCurrency.add(
            invoiceItem.CtAccountNo,
            invoiceItem.Currency,
          );
        } else {
          //update the credit account amount
          const d = htCreditAccountAmount.get(invoiceItem.CtAccountNo);
          htCreditAccountAmount.add(
            invoiceItem.CtAccountNo,
            d + invoiceItem.Amount,
          );
        }
      });

      const savedItems = htCreditAccountAmount.list();

      for (let i = 0; i < savedItems.length; i++) {
        const journalEntry = new JournalEntry(dbTransaction);
        //Temporary fix to successfully save journal entry in PostJournal
        journalEntry.init({
          CompanyId: this.CompanyId,
          Name: 'issue Invoice ' + invoice.DocNo,
        });
        const account = await Account.initAccount(
          dbTransaction,
          savedItems[i].value[0],
        );
        const creditAmount = savedItems[i].value[1];
        const currency = htCreditAccountCurrency.get(savedItems[i].value[0]);

        const dt = await journalEntry.newLedgerTransaction(
          TransactionTypeOptions.DEBIT,
        );

        if (dtAccountNo) {
          /*Transacting using AR account provided*/
          dt.AccountNo = dtAccountNo;
        } else {
          /*Transacting based on default customer AR account*/
          dt.AccountNo = (await customer.AccountReceivable).AccountNo;
        }

        dt.Currency = currency ? currency : 'MYR';
        dt.DebitAmount = creditAmount ? creditAmount : 0.0;
        dt.Date = transactionDate;
        dt.Description = account.Name;
        dt.Name = account.Name;
        dt.RelatedDocNo = invoice.DocNo;

        const ct = await journalEntry.newLedgerTransaction(
          TransactionTypeOptions.CREDIT,
        );
        ct.AccountNo = savedItems[i].key;
        ct.Currency = currency ? currency : 'MYR';
        ct.CreditAmount = creditAmount ? creditAmount : 0.0;
        ct.Date = transactionDate;
        ct.Description = account.Name;
        ct.Name = account.Name;
        ct.RelatedDocNo = invoice.DocNo;

        await this.postJournal(dbTransaction, journalEntry);
      }

      return invoice;
    } catch (err) {
      // tslint:disable-next-line:no-console
      console.log('Issue debit note err: ', err);
      throw err;
    }
  }

  /**
   * Issue a credit note
   *
   * @param dbTransaction - The database transaction to be used
   * @param loginUser - The user issuing the credit note
   * @param creditNote - The document containing the credit note details
   * @param customer - The customer to be issued the credit note
   * @param ctAccountNo - The account number of the Account Payable (AP) account to debit. If not provided, will debit the customer's default AR account
   *
   * @returns {Document} - Document object representing the full details of the invoice issued
   */
  async issueCreditNote(
    dbTransaction: any,
    loginUser: LoginUserBase,
    creditNote: Document,
    customer: FinanceCustomerBase,
    ctAccountNo?: string,
  ): Promise<Document> {
    try {
      /*Check if the invoice number already exists (should be unique)*/
      const duplicateCreditNote =
        await FinanceCompany._DocumentRepository.findOne({
          where: {
            DocNo: creditNote.DocNo,
          },
          transaction: dbTransaction,
        });

      if (duplicateCreditNote) {
        throw new Error('Invoice number already exists');
      }

      const documentItems = await creditNote.DocumentItems;

      /*Check if the document has at least 1 document item*/
      if (!documentItems.length) {
        throw new Error('Document must have at least 1 document item');
      }

      /*Check if each document item has CtAccountNo provided*/
      for (const invoiceItem of documentItems) {
        if (!invoiceItem.CtAccountNo) {
          throw new Error(
            'Each document item should have CtAccountNo provided',
          );
        }
      }

      /*Set up the document type*/
      creditNote.DocType = DocType.DEBIT_NOTE;

      /*Saving the document and document items to the database*/
      await FinanceCompany._DocumentRepository.create(
        {
          DocNo: creditNote.DocNo,
          DocType: creditNote.DocType,
          DocDate: creditNote.DocDate,
          CompanyId: creditNote.CompanyId,
          Currency: creditNote.Currency,
          Amount: creditNote.Amount,
          Description: creditNote.Description,
          Status: creditNote.Status,
          IssuedById: loginUser.ObjectId,
          IssuedToId: customer.ObjectId,
          IssuedToType: creditNote.IssuedToType,
          RelatedObjectId: creditNote.RelatedObjectId,
          RelatedObjectType: creditNote.RelatedObjectType,
          CreatedById: loginUser.ObjectId,
          CreatedAt: new Date(),
          UpdatedById: loginUser.ObjectId,
          UpdatedAt: new Date(),
          DocPDFFileMediaId: creditNote.DocPDFFileMediaId,
          DocHTMLFileMediaId: creditNote.DocHTMLFileMediaId,
          AccSystemRefId: creditNote.AccSystemRefId,
          PostedToAccSystemYN: creditNote.PostedToAccSystemYN,
          PostedById: creditNote.PostedById,
          PostedDateTime: new Date(),
          UseAccSystemDocYN: creditNote.UseAccSystemDocYN,
        },
        {
          transaction: dbTransaction,
        },
      );

      documentItems.forEach(async (documentItem) => {
        await FinanceCompany._DocumentItemRepository.create(
          {
            DocumentItemId: documentItem.DocumentItemId,
            DocNo: documentItem.DocNo,
            Name: documentItem.Name,
            NameBM: documentItem.NameBM,
            Description: documentItem.Description,
            ItemId: documentItem.ItemId,
            ItemType: documentItem.ItemType,
            ItemSKU: documentItem.ItemSKU,
            ItemSerialNo: documentItem.ItemSerialNo,
            Currency: documentItem.Currency,
            UnitPrice: documentItem.UnitPrice,
            Quantity: documentItem.Quantity,
            QuantityUOM: documentItem.QuantityUOM,
            Amount: documentItem.Amount,
            TaxCode: documentItem.TaxCode,
            TaxAmount: documentItem.TaxAmount,
            TaxRate: documentItem.TaxRate,
            TaxInclusiveYN: documentItem.TaxInclusiveYN,
            DtAccountNo: documentItem.DtAccountNo,
            CtAccountNo: documentItem.CtAccountNo,
          },
          {
            transaction: dbTransaction,
          },
        );
      });

      /*Generating the credit note*/
      if (creditNote.UseAccSystemDocYN === 'Y') {
        /*todo: Posting to accounting system to generate creditNote*/
        await this.AccountingSystem.createCreditNote(creditNote);
      } else {
        /*todo: check config file to see which invoice template is to be used for specific project*/

        /*Generating credit note based on template*/
        creditNote.generateCreditNote(loginUser.IDNo, customer);
      }

      const journalEntry = new JournalEntry(dbTransaction);
      //Temporary fix to successfully save journal entry in PostJournal
      journalEntry.init({
        CompanyId: this.CompanyId,
        Name: 'Issue Debit Note ' + creditNote.DocNo,
      });

      const transactionDate = new Date();

      const creditTransaction = await journalEntry.newLedgerTransaction(
        TransactionTypeOptions.CREDIT,
      );

      if (ctAccountNo) {
        /*Transacting using AR account provided*/
        creditTransaction.AccountNo = ctAccountNo;
      } else {
        /*Transacting based on default customer AR account*/
        creditTransaction.AccountNo = customer.CustSystemCode + '-AP';
      }

      creditTransaction.Currency = creditNote.Currency;
      creditTransaction.CreditAmount = creditNote.Amount;
      creditTransaction.Date = transactionDate;
      creditTransaction.Description = creditNote.DocNo;
      creditTransaction.Name = creditNote.DocNo;

      for (const invoiceItem of documentItems) {
        const itemLedger = await journalEntry.newLedgerTransaction(
          TransactionTypeOptions.DEBIT,
        );
        itemLedger.AccountNo = invoiceItem.CtAccountNo;
        itemLedger.Currency = invoiceItem.Currency;
        itemLedger.DebitAmount = invoiceItem.Amount;
        itemLedger.Date = transactionDate;
        itemLedger.Description = invoiceItem.Description;
        itemLedger.Name = invoiceItem.Name;
      }

      this.postJournal(dbTransaction, journalEntry);

      const payload = {
        Action: 'Create',
        Activity: 'Issuing a Credit Note Transaction',
        Description: `Credit Transaction (ID: ${creditTransaction.TransactionId}) has been created`,
        EntityType: 'CreditTransaction',
        EntityValueBefore: JSON.stringify({}),
        EntityValueAfter: JSON.stringify(creditTransaction),
        PerformedById: 'test',
        PerformedAt: transactionDate,
        EntityId: creditTransaction.TransactionId,
      };

      await axios.post(
        `${process.env.COMMON_API_URL}/activity-histories`,
        payload,
      );

      return creditNote;
    } catch (err) {
      // tslint:disable-next-line:no-console
      console.log('Issue credit note err: ', err);
      throw err;
    }
  }

  /**
   * Register a payment collected
   *
   * @param dbTransaction - The database transaction to be used
   * @param loginUser - The user collecting the payment
   * @param payment - The payment object containing payment details
   * @param customer - The customer making the payment
   * @param ctAccountNo - The account number of the customer's Account Receivable account to transact, else the default customer account payable receivable will be used
   *
   * @returns {Payment} - Payment object representing the full detals of the payment collection recorded
   */
  async collectPayment(
    dbTransaction: any,
    loginUser: LoginUserBase,
    payment: Payment,
    customer: FinanceCustomerBase,
    ctAccountNo?: string,
  ): Promise<Payment> {
    try {
      /*validation 1: Make sure that the payment has at least 1 payment item*/
      const paymentItems = await payment.PaymentItems;
      if (paymentItems.length < 1) {
        throw new Error(
          'Atleast one payment item is required to identify what payment is being paid for.',
        );
      }

      /*validation 2: Make sure that the payment has at least 1 payment method*/
      const paymentPaidWithItems = await payment.PaymentPaidWith;
      if (paymentPaidWithItems.length < 1) {
        throw new Error(
          'Atleast one payment paid with item is required to identify how the payment was made.',
        );
      }

      payment.PaymentId = cuid();
      payment.PaymentType = PaymentType.PAYMENT_RECEIVED;
      payment.ReceivedBy = loginUser.ObjectId;

      await FinanceCompany._PaymentRepository.create(
        {
          PaymentId: payment.PaymentId,
          PaymentType: payment.PaymentType,
          PaymentDate: payment.PaymentDate,
          Description: payment.Description,
          Currency: payment.Currency,
          Amount: payment.Amount,
          Status: PaymentStatus.SUCCESSFUL,
          PostedToAccSystemYN: 'N',
          ReceivedBy: payment.ReceivedBy,
          UpdatedAt: new Date(),
          UpdatedBy: loginUser.ObjectId,
          CreatedAt: new Date(),
          CreatedBy: loginUser.ObjectId,
        },
        { transaction: dbTransaction },
      );

      for (const paymentItem of paymentItems) {
        await FinanceCompany._PaymentItemRepository.create(
          {
            PaymentId: payment.PaymentId,
            PayForObjectId: paymentItem.PayForObjectId,
            PayForObjectType: paymentItem.PayForObjectType,
            Currency: paymentItem.Currency,
            Amount: paymentItem.Amount,
            Name: paymentItem.Name,
            Description: paymentItem.Description,
          },
          { transaction: dbTransaction },
        );
      }

      for (const paymentPaidWithItem of paymentPaidWithItems) {
        await FinanceCompany._PaymentPaidWithRepository.create(
          {
            PaymentId: payment.PaymentId,
            MethodTypeId: paymentPaidWithItem.MethodTypeId,
            Currency: paymentPaidWithItem.Currency,
            Amount: paymentPaidWithItem.Amount,
            Status: paymentPaidWithItem.Status,
            TransactionId: paymentPaidWithItem.TransactionId,
            RefBank: paymentPaidWithItem.RefBank,
            RefName: paymentPaidWithItem.RefName,
            RefNo: paymentPaidWithItem.RefNo,
            RefOther1: paymentPaidWithItem.RefOther1,
            RefOther2: paymentPaidWithItem.RefOther2,
            RefOther3: paymentPaidWithItem.RefOther3,
            RefOther4: paymentPaidWithItem.RefOther4,
            RefOther5: paymentPaidWithItem.RefOther5,
            Remarks: paymentPaidWithItem.Remarks,
            PaymentMediaId: paymentPaidWithItem.PaymentMediaId,
          },
          { transaction: dbTransaction },
        );
      }

      /*Registering the Journal Entries for the transaction*/
      const transactionDate = new Date();

      for (const paymentPaidWith of paymentPaidWithItems) {
        let paymentMethodType: PaymentMethodType;

        try {
          paymentMethodType = new PaymentMethodType(
            dbTransaction,
            paymentPaidWith.MethodTypeId,
          );
        } catch (error) {
          if (error instanceof RecordNotFoundError) {
            throw new Error('Invalid Payment Method Type Id');
          } else {
            throw error;
          }
        }
        const journalEntry = new JournalEntry(dbTransaction);
        //Temporary fix to successfully save journal entry in PostJournal
        journalEntry.init({
          CompanyId: this.CompanyId,
          Name: 'Collect-payments for ' + payment.PaymentId,
        });

        const debitLT = await journalEntry.newLedgerTransaction(
          TransactionTypeOptions.DEBIT,
        );
        debitLT.AccountNo = paymentMethodType.AccountNo;
        debitLT.Currency = payment.Currency;
        debitLT.DebitAmount = payment.Amount;
        debitLT.Date = transactionDate;
        debitLT.Description = customer.FullName;
        debitLT.Name = customer.FullName;

        const creditLT = await journalEntry.newLedgerTransaction(
          TransactionTypeOptions.CREDIT,
        );

        if (ctAccountNo) {
          creditLT.AccountNo = ctAccountNo;
        } else {
          creditLT.AccountNo = (await customer.AccountReceivable).AccountNo;
        }
        creditLT.Currency = payment.Currency;
        creditLT.CreditAmount = payment.Amount;
        creditLT.Date = transactionDate;
        creditLT.Description = paymentMethodType.Name;
        creditLT.Name = paymentMethodType.Name;

        await this.postJournal(dbTransaction, journalEntry);
      }

      /*todo: saving a record into the activity history table*/

      return payment;
    } catch (error) {
      throw error;
    }
  }

  /**
   * Method to make payment to a customer.
   *
   * @param dbTransaction - The database transaction to be used
   * @param loginUser - The user logging in and the user who collected the payment.
   * @param payment - The payment object containing payment details
   * @param customer - The customer who is receiving the payment.
   * @param ctAccountNo - The account number of the customer's Account Payable account to transact, else the default customer account payable will be used.
   *
   * @returns {Payment} - Payment object representing the full details of the payment collection recorded
   */
  async makePayment(
    dbTransaction: any,
    loginUser: LoginUserBase,
    payment: Payment,
    customer: FinanceCustomerBase,
    dtAccountNo?: string,
  ): Promise<Payment> {
    let paymentMethodType: PaymentMethodType;
    try {
      paymentMethodType = new PaymentMethodType(dbTransaction);

      const paymentItems = await payment.PaymentItems;
      if (paymentItems.length < 1) {
        throw new Error(
          'Atleast one payment item is required to identify what payment is being paid for',
        );
      }

      payment.PaymentId = cuid();
      payment.PaymentType = PaymentType.PAYOUT;
      payment.ReceivedBy = loginUser.ObjectId;

      await FinanceCompany._PaymentRepository.create(
        {
          PaymentId: payment.PaymentId,
          PaymentType: payment.PaymentType,
          PaymentDate: payment.PaymentDate,
          Description: payment.Description,
          Currency: payment.Currency,
          ReceivedBy: payment.ReceivedBy,
          Amount: payment.Amount,
          Status: PaymentStatus.SUCCESSFUL,
          PostedToAccSystemYN: 'N',
          UpdatedAt: new Date(),
          UpdatedBy: loginUser.ObjectId,
          CreatedAt: new Date(),
          CreatedBy: loginUser.ObjectId,
        },
        { transaction: dbTransaction },
      );

      paymentItems.forEach(async (paymentItem) => {
        await FinanceCompany._PaymentItemRepository.create(
          {
            PaymentId: payment.PaymentId,
            PayForObjectId: paymentItem.PayForObjectId,
            PayForObjectType: paymentItem.PayForObjectType,
            Currency: paymentItem.Currency,
            Amount: paymentItem.Amount,
            Name: paymentItem.Name,
            Description: paymentItem.Description,
          },
          { transaction: dbTransaction },
        );
      });

      const journalEntry = new JournalEntry(dbTransaction);
      //Temporary fix to successfully save journal entry in PostJournal
      journalEntry.init({
        CompanyId: this.CompanyId,
        Name: 'Make Payment for ' + payment.PaymentId,
      });
      const transactionDate = new Date();

      const creditLT = await journalEntry.newLedgerTransaction(
        TransactionTypeOptions.CREDIT,
      );
      creditLT.AccountNo = paymentMethodType.AccountNo;
      creditLT.Currency = payment.Currency;
      creditLT.CreditAmount = payment.Amount;
      creditLT.Date = transactionDate;
      creditLT.Description = customer.FullName;
      creditLT.Name = customer.FullName;

      const debitLT = await journalEntry.newLedgerTransaction(
        TransactionTypeOptions.DEBIT,
      );
      if (dtAccountNo) {
        debitLT.AccountNo = dtAccountNo;
      } else {
        debitLT.AccountNo = (await customer.AccountPayable).AccountNo;
      }
      debitLT.Currency = payment.Currency;
      debitLT.DebitAmount = payment.Amount;
      debitLT.Date = transactionDate;
      debitLT.Description = paymentMethodType.Name;
      debitLT.Name = paymentMethodType.Name;

      await this.postJournal(dbTransaction, journalEntry);

      /*todo: saving a record into the activity history table*/

      return payment;
    } catch (error) {
      if (error instanceof RecordNotFoundError) {
        throw new Error('Invalid PaymentMethodType id');
      } else {
        throw error;
      }
    }
  }

  get PaymentMethods(): Promise<PaymentMethod[]> {
    return new Promise((resolve, reject) => {
      if (this.CompanyId !== 'New') {
        FinanceCompany._PaymentMethodRepository
          .findAll({
            where: {
              CompanyId: this.CompanyId,
            },
            transaction: this._DbTransaction,
          })
          .then((paymentMethod) => {
            const paymentMethodObjects = paymentMethod.map(
              (paymentMethodData) => {
                return new Promise((resolve, reject) => {
                  FinanceCompany._PaymentMethodTypeRepository
                    .findAll({
                      where: {
                        MethodId: paymentMethodData.MethodId,
                      },
                      transaction: this._DbTransaction,
                      raw: true,
                    })
                    .then((paymentMethodTypes) => {
                      const paymentMethodObjects = {
                        ...paymentMethodData.get({ plain: true }),
                        Types: paymentMethodTypes,
                      };
                      resolve(paymentMethodObjects);
                    })
                    .catch((err) => {
                      reject(err);
                    });
                }).then((paymentMethods) => paymentMethods);
              },
            );
            return Promise.all(paymentMethodObjects);
          })
          .then((paymentMethodObjects) => {
            this._PaymentMethods = paymentMethodObjects;
            resolve(this._PaymentMethods);
          })
          .catch((err) => {
            reject(err);
          });
      } else {
        resolve(this._PaymentMethods);
      }
    });
  }

  /**
   * Method to load / reload the payment methods based on the configuration file
   */
  async LoadPaymentMethods(): Promise<void> {
    /*Retrieve the payment method from the config file */
    const { companyId, paymentMethods } = config.financeCompanies['TXG-FS'];
    const paymentMethod = new PaymentMethod();
    for (const method in paymentMethods) {
      const paymentMethodData =
        await FinanceCompany._PaymentMethodRepository.findOne({
          where: {
            MethodId: paymentMethods[method].id,
            Name: paymentMethods[method].name,
          },
        });

      if (!paymentMethodData) {
        const newPaymentMethod =
          await FinanceCompany._PaymentMethodRepository.create({
            MethodId: paymentMethods[method].id,
            Name: paymentMethods[method].name,
            CompanyId: companyId,
          });

        this._PaymentMethods.push(newPaymentMethod);
      }
      this._PaymentMethods.push(paymentMethodData);
    }

    this._PaymentMethods.forEach(async (item) => {
      const p = item?.get({ plain: true });

      for (const method in paymentMethods) {
        if (!p) {
          continue;
        }

        if (p.MethodId === paymentMethods[method]?.id) {
          const paymentMethodTypeData =
            await FinanceCompany._PaymentMethodTypeRepository.findOne({
              where: {
                MethodId: p.MethodId,
              },
            });

          if (!paymentMethodTypeData) {
            const configPaymentMethodTypes = paymentMethods[method]?.types;

            for (const methodType in configPaymentMethodTypes) {
              // TODO: Create a seeder for payment method account
              /*validate whether account data already exists*/
              const accountData =
                await FinanceCompany._AccountRepository.findOne({
                  where: {
                    AccountNo: configPaymentMethodTypes[methodType].accountNo,
                  },
                });

              /*generating account data if not exist */
              if (!accountData) {
                const accountPayload = {
                  CompanyId: companyId,
                  Name: configPaymentMethodTypes[methodType].name,
                  AccountType: 'PaymentMethod',
                  CreatedAt: new Date(),
                  CreatedById: 'System',
                  AccSystemRefId: 'REF',
                  PostedToAccSystemYN: 'N',
                };

                try {
                  await FinanceCompany._AccountRepository.create({
                    AccountNo: configPaymentMethodTypes[methodType].accountNo,
                    ...accountPayload,
                  });

                  await FinanceCompany._AccountRepository.create({
                    AccountNo:
                      configPaymentMethodTypes[methodType]
                        .processingFeeAccountNo,
                    ...accountPayload,
                  });
                } catch (err) {}
              }

              const paymentMethodTypePayload = {
                MethodId: p.MethodId,
                MethodTypeId: configPaymentMethodTypes[methodType].id,
                Name: configPaymentMethodTypes[methodType].name,
                AccountNo: configPaymentMethodTypes[methodType].accountNo,
                ProcessingFeeRate:
                  configPaymentMethodTypes[methodType].processingFeeRate,
                ProcessingFeeAccountNo:
                  configPaymentMethodTypes[methodType].processingFeeAccountNo,
              };

              try {
                await paymentMethod.newPaymentMethodType(
                  paymentMethodTypePayload,
                );
              } catch (err) {}
            }
          }
        }
      }
    });
  }
}
