import { ObjectBase, RecordNotFoundError } from '@tomei/general';
import { PaymentType } from '../enum';
import { IPaymentParams } from './interfaces/payment-params.interface';
import { AccountSystemEntity } from '../account-system-entity/account-system-entity';
import { PaymentStatus } from '../enum/payment-status.enum';
import { PaymentRepository } from './payment.repository';
import { PaymentItemRepository } from '../payment-item/payment-item.repository';
// import PaymentItemModel from '../models/payment-item.entity';
import PaymentItem from '../payment-item/payment-item';
import PaymentPaidWith from '../payment-paid-with/payment-paid-with';
import { PaymentPaidWithRepository } from '../payment-paid-with/payment-paid-with.repository';

export default class Payment extends AccountSystemEntity {
  private _PaymentId = 'New';
  PaymentType: PaymentType;
  PaymentDate: Date;
  Description = '';
  Currency = 'MYR';
  Amount = 0;
  Status: PaymentStatus;
  ReceivedBy = '';
  UpdatedBy = '';
  UpdatedAt: Date;
  Remarks = '';

  private static _RepositoryBase = new PaymentRepository();
  private static _PaymentItemRepository = new PaymentItemRepository();
  private static _PaymentPaidWithRepository = new PaymentPaidWithRepository();

  private _DbTransaction: any;
  private _PaymentItems = [];
  private _PaymentPaidWith = [];

  paymentItemRepository: PaymentItemRepository;

  public get PaymentId() {
    return this._PaymentId;
  }

  public set PaymentId(id: string) {
    this._PaymentId = id;
  }

  /* AccountSystemEntity Implementation */
  get RepositoryBase() {
    return Payment._RepositoryBase;
  }

  /* ObjectBase Implementation */
  get ObjectId() {
    return this.PaymentId;
  }

  get ObjectName() {
    return this.PaymentId;
  }

  get TableName() {
    return 'finance_Payment';
  }

  constructor(dbTransaction?: any, paymentId?: string) {
    super();
    if (dbTransaction) {
      this._DbTransaction = dbTransaction;
    }
    if (paymentId) {
      this.RepositoryBase.findOne({
        where: {
          PaymentId: paymentId,
        },
        transaction: dbTransaction,
      })
        .then((paymentData) => {
          if (paymentData) {
            this.PaymentType = paymentData.PaymentType;
            this.PaymentDate = paymentData.PaymentDate;
            this.Description = paymentData.Description;
            this.Currency = paymentData.Currency;
            this.Amount = paymentData.Amount;
            this.Status = paymentData.Status;
            this.ReceivedBy = paymentData.ReceivedBy;
            this.UpdatedAt = paymentData.UpdatedAt;
            this.UpdatedBy = paymentData.UpdatedBy;
            this.Remarks = paymentData.Remarks;
            this.AccSystemRefId = paymentData.AccSystemRefId;
            this.PostedToAccSystemYN = paymentData.PostedToAccSystemYN;
            this.PostedById = paymentData.PostedById;
            this.PostedDateTime = paymentData.PostedDateTime;
          } else {
            const notFoundError = new RecordNotFoundError('No Record Found.');
            throw notFoundError;
          }
        })
        .catch((err) => {
          // tslint:disable-next-line:no-console
          console.log('Payment Class constructor err: ', err);
        });
    }
  }

  // public get PaymentItems() {
  //   if (this._PaymentItems === null) {
  //     this._PaymentItems = new Array<PaymentItemModel>();
  //     if (this.PaymentId !== 'New') {
  //       this.paymentItemRepository
  //         .findAll({
  //           where: {
  //             PaymentId: this.PaymentId,
  //           },
  //         })
  //         .then((allPaymentItems) => {
  //           this._PaymentItems = allPaymentItems;
  //         });
  //     }
  //   }

  //   return this._PaymentItems;
  // }

  get PaymentItems(): Promise<PaymentItem[]> {
    return new Promise((resolve, reject) => {
      if (this.PaymentId !== 'New') {
        Payment._PaymentItemRepository
          .findAll({
            where: {
              PaymentId: this.PaymentId,
            },
            transaction: this._DbTransaction,
          })
          .then((paymentItems) => {
            const paymentItemObjects = paymentItems.map((paymentItem) => {
              new PaymentItem(
                new Payment(this._DbTransaction, paymentItem.PaymentId),
                {
                  ObjectId: this.ObjectId,
                  ObjectName: this.ObjectName,
                  TableName: this.TableName,
                },
              );
            });
            return Promise.all(paymentItemObjects);
          })
          .then((paymentItemObjects) => {
            this._PaymentItems = paymentItemObjects;
            resolve(this._PaymentItems);
          })
          .catch((err) => {
            reject(err);
          });
      }
      resolve(this._PaymentItems);
    });
  }

  get PaymentPaidWith(): Promise<PaymentPaidWith[]> {
    return new Promise((resolve, reject) => {
      if (this.PaymentId !== 'New') {
        Payment._PaymentPaidWithRepository
          .findAll({
            where: {
              PaymentId: this.PaymentId,
            },
            transaction: this._DbTransaction,
          })
          .then((paymentPaidWithItems) => {
            const paymentPaidWithObjects = paymentPaidWithItems.map(
              (paymentPaidWithItem) => {
                new PaymentPaidWith(
                  new Payment(
                    this._DbTransaction,
                    paymentPaidWithItem.PaymentId,
                  ),
                  paymentPaidWithItem.MethodTypeId,
                );
              },
            );
            return Promise.all(paymentPaidWithObjects);
          })
          .then((paymentPaidWithObjects) => {
            this._PaymentPaidWith = paymentPaidWithObjects;
            resolve(this._PaymentPaidWith);
          })
          .catch((err) => {
            reject(err);
          });
      }
      resolve(this._PaymentPaidWith);
    });
  }

  newPaymentItem(objectBeingPaidFor: ObjectBase): PaymentItem {
    const paymentItem = new PaymentItem(this, objectBeingPaidFor);
    this._PaymentItems.push(paymentItem);
    return paymentItem;
  }

  newPaymentPaidWith(paymentMethodTypeId: string) {
    const paymentPaidWith = new PaymentPaidWith(this, paymentMethodTypeId);
    this._PaymentPaidWith.push(paymentPaidWith);
    return paymentPaidWith;
  }

  init(Params: IPaymentParams): void {
    if (Params.Currency) this.Currency = Params.Currency;
    if (Params.PaymentType) this.PaymentType = Params.PaymentType;
  }
}
